!=============================================================================================================
! ControlMonitor Class          ( ControlMonitor.clw )
! Version 2.4
! Last Updated: 04-12-2004
!
!=============================================================================================================
! This template/class set was assembled by Charles Edmonds and is based on code and
! much help from Arnor Baldvinsson.  It also contains suggestions from other members
! of the Clarion Development Community.
!
! Thanks to all who contributed!
!=============================================================================================================
! Background
!
! This class / template set was written to provide a better way to execute code when the
! the contents of ANY field or control was changed - either by direct user interaction or
! by a direct code assignment (such as Variable=Value).  The goal was to eliminate the need
! to call a routine or enter code at each embed point on a screen with many controls.
!
! The idea for the set (and much of the initial code) came from Arnor Baldvinsson in response
! to an inquiry in the Clarion Newsgroups by Charles Edmonds.  Charles was responsible for
! coding the actual class structure and then implementing the template to make it easy to use.
!
! Support was added for the Sysinternals DebugView program so that a programmer could view output
! as the class was working without relying on STOP, HALT or MESSAGE statements (which are known
! to interfer with the code processing.
!---------------------------------------------------------------------------------------------
! Class Dynamics
!
! The class works by building a queue of the FEQ numbers and a CRC for the contents of each control 
! as the window is initialized.  Then when a Field Event is generated or an EVENT:Accepted is posted,
! the class checks the CRC of all controls against the data in the queue.  If the CRC's do not match
! then any code the programmer has placed in a ROUTINE that is generated by the template is called.
!
! In the case of a Field Event (such as a entry, spin control, combo, etc.) - in other words a
! direct action as the result of the user interacting with the program, the "HasChanged" method
! is called and the contents of the specific field are compared against the record in the queue.
! IF they do not match, the queue is updated with the new CRC and the code in the routine is called.
!
! In the case of an EVENT:Accepted, the "AnyChanged" method of the class is called and the contents
! of each control is compared against the value stored in the queue.  If they do not match, then the
! queue record is updated with the new CRC and the code in the routine is called.
!
!
! This dual approach method makes sure that the code in the routine is ALWAYS called if a change in
! the data occurs - whether by user interaction with the program, or by direct code assignment.
!---------------------------------------------------------------------------------------------
! Revisions
! 1.1   Changed class to trap field events and Event:Accepted
! 1.2   Added support for SysInternals Debugger
! 1.3   Changed to CRC method of verifying data
! 1.4   Localized instances of the class
! 1.5   Updated template to allow programmers to name class instances
! 2.0   Added Extended CASE statement generation and support for multi-DLL applications
! 2.1   Changed name of embed points in Extended CASE for clarity
! 2.2   Added a CLEAR( %classname.lastchanged ) after the CASE statement to avoid duplicate code calls
! 2.3   Added a checkbox on the Global Extension tab to disable ALL output to DebugView
!       (This makes it easy to be sure all output is turned OFF when you ship your program)
! 2.4   Added support for LIST controls in Extended CASE template
!---------------------------------------------------------------------------------------------
! Instructions
!
! 1.    Add the ControlMonitor Global Extension Template to the Global section of your program.
! 2.    Add the ControlMonitor Detect Template to each procedure where you want to use it.
! 3.    Set the name for the localized instance of the class (or accept the default value)
! 4.    If you want to output to the Sysinternals DebugView program, check the box on the extension template.
!
!       You can download the debugger from www.sysinternals.com if you do not have it.
!       Run the DBGVIEW.EXE executable before you start your program to see the output.
!       When you no longer need it, just uncheck the box on the template extension tab.
!
!       Note:  You can also set the output to the debugger ON/OFF via code by setting the following
!              value in your code AFTER the INIT function has been called.
!
!              classinstancename.wDebugAll = True  (or False)
!---------------------------------------------------------------------------------------------
! Changes / Additions
!
! If you make useful changes to this template or class set, please email a copy to Charles Edmonds
! at "ControlMonitor@lansrad.com" so we can update the copies at the download centers
!---------------------------------------------------------------------------------------------

    MEMBER

    MAP
      !Needed for Sysinternals Debugger
      Module('win32.lib')
    ControlMonitorOutputDebugString(*CSTRING),PASCAL,RAW,NAME('OutputDebugStringA')
      End

      MODULE('CLARION 32 BIT API')
    ControlMonitorCRC32(*CSTRING buffer,ULONG size,ULONG crc),ULONG,RAW,NAME('CLA$CRC32')
      END !module

      MODULE('C55Run')
    ControlMonitorFieldName (SIGNED feq), *CSTRING, RAW, NAME('Cla$FIELDNAME')
      END

    END

    INCLUDE('ControlMonitor.inc'),ONCE




CQ                      QUEUE,TYPE
FEQ                 LONG
ContCRC             LONG
FEQLabel            CSTRING(256)
                        END


!===============================================================================================
! ControlMonitor:  A CLASS to always detect changes in data for any field on a window
!
! Class Methods
!===============================================================================================

!-----------------------------------------------------------------------------------------------
! Init the class
!-----------------------------------------------------------------------------------------------
ControlMonitor.Init    Procedure
I   LONG
 Code
 Self.Controls   &= New CQ
 Loop I = FirstField() To LastField()
   Self.Controls.FEQ      = I                               !Get the FEQ
   Self.Controls.ContCRC = Self.GetCRC( I )                 !Get the CRC
   Self.Controls.FEQLabel = ControlMonitorFieldName( I )    !Get the field label for the debugger
   Add(Self.Controls,Self.Controls.FEQ)
 End
 Self.Initialized = True


!-----------------------------------------------------------------------------------------------
! Kill the class
!-----------------------------------------------------------------------------------------------
ControlMonitor.Kill    Procedure
 Code
 Free(Self.Controls)
 Dispose(Self.Controls)
 Self.Initialized = False


!-----------------------------------------------------------------------------------------------
! Determine if the control being checked is currently selected
!-----------------------------------------------------------------------------------------------
ControlMonitor.IsSelected      Procedure(Long pFEQ)
Found  Byte
 Code
 If Self.Initialized
   Self.Controls.Feq = pFEQ
   Get(Self.Controls,Self.Controls.Feq)
   If ErrorCode()
     Found = False
   Else
     Found = True
   End
 End
 Return(Found)


!-----------------------------------------------------------------------------------------------
! Determine if the contents of a specific control has changed
!-----------------------------------------------------------------------------------------------
ControlMonitor.HasChanged     Procedure(Long pFEQ)
Changed Byte,Auto
LocalCRC    LONG

 Code
 Changed = False
 LocalCRC = 0
 Self.Controls.FEQ = pFEQ
 Get(Self.Controls,Self.Controls.Feq)
 If Not ErrorCode()
   LocalCRC = Self.GetCRC(Self.Controls.Feq)  !Get the CRC of this field
   If Self.Controls.ContCRC <> LocalCRC
     !------------------------------------------------------------------------------------
     !Output message to SysInternals Debugger
     IF Self.wDebugAll = True
       Self.wDebug('CM.HasChanged: ' & CLIP(Self.Controls.FEQLabel) & ' Old CRC[' & |
         CLIP(Self.Controls.ContCRC) & '] New CRC[' & LocalCRC & ']')
     END
     !------------------------------------------------------------------------------------
     Self.Controls.ContCRC = LocalCRC       !Update the new CRC value to the queue
     Put(Self.Controls)                           
     Changed = True
     !-------------------------------------------
     !Revision 2.0 - 8-11-03, Charles Edmonds
     !-------------------------------------------
     !Update property so we can tell what changed
     SELF.LastChanged = Self.Controls.Feq

     !Output message to SysInternals Debugger
     IF Self.wDebugAll = True
       Self.wDebug('Last Control Changed: ' & Self.Controls.Feq)
     End !IF
     !-------------------------------------------
   End
 End
 Return(Changed)
                                                            
                     
!-----------------------------------------------------------------------------------------------
! Determine if the contents of ANY control has changed
!-----------------------------------------------------------------------------------------------
ControlMonitor.AnyChanged     Procedure
Changed Byte,Auto
Cnt     LONG
LocalCRC    LONG
 Code
 Changed = False
 LocalCRC = 0

 LOOP Cnt = 1 TO RECORDS( Self.Controls )
   Get(Self.Controls,Cnt)
   If Not ErrorCode()
     LocalCRC = Self.GetCRC(Self.Controls.Feq)  !Get the CRC of this field
     If Self.Controls.ContCRC <> LocalCRC
       !------------------------------------------------------------------------------------
       !Output message to SysInternals Debugger
       IF Self.wDebugAll = True
         Self.wDebug('CM.AnyChanged: ' & CLIP(Self.Controls.FEQLabel) & ' Old CRC[' & |
           CLIP(Self.Controls.ContCRC) & '] New CRC[' & LocalCRC & ']')
       END
       !------------------------------------------------------------------------------------
       Self.Controls.ContCRC = LocalCRC       !Update the new CRC value to the queue
       Put(Self.Controls)
       Changed = True
       !-------------------------------------------
       !Revision 2.0 - 8-11-03, Charles Edmonds
       !-------------------------------------------
       !Update property so we can tell what changed
       SELF.LastChanged = Self.Controls.Feq

       !Output message to SysInternals Debugger
       IF Self.wDebugAll = True
         Self.wDebug('Last Control Changed: ' & Self.Controls.Feq)
       End !IF
       !-------------------------------------------
     End
   End
 END
 Return(Changed)


!-----------------------------------------------------------------------------------------------
! Add support for Sysinternals Debugger
!------------------------------------------------------------------------------------------------
ControlMonitor.wDebug              Procedure(String pS)
!------------------------------------------------------------------------------------------------
! This outputs the string to DebugView:
! DebugView
! Copyright (C) 1998-2000 Mark Russinovich 
! Sysinternals - www.sysinternals.com
! Executable:  DebugView.exe
! Run the program before using this statement and the output will appear in the program.
s  CString(1025)
 Code
 !Format the string
 s = Format(Clock(),@t04) & ' ' & Clip(pS)
 !Output to DebugView
 ControlMonitorOutputDebugString(s)


!-----------------------------------------------------------------------------------------------
! Get the CRC for the data in the field to be examined
!-----------------------------------------------------------------------------------------------
ControlMonitor.GetCRC     Procedure(Long pFEQ)

TheCString  &CSTRING
TheSize     LONG
CRCValue    LONG

 Code

   !Get the size of the current control's contents
   TheSize = LEN(Contents(pFEQ))

   !Allocate a new string at the new max size
   TheCString &= NEW(CSTRING(TheSize + 1))       !Add 1 because the CRC uses a STRING

   !Get the Data in this control
   TheCString = CONTENTS( pFEQ )

   !Get the CRC
   CRCValue = ControlMonitorCRC32( TheCString, SIZE(TheCString), 0 )

   !Dispose of the string
   DISPOSE(TheCString)

   !Return the new CRC value
   RETURN( CRCValue )


!-----------------------------------------------------------------------------------------------
! Remove a control from the queue of controls to be checked
!-----------------------------------------------------------------------------------------------
ControlMonitor.RemoveControl  Procedure(Long pFEQ)

 Code
 Self.Controls.Feq = pFEQ
 Get(Self.Controls,Self.Controls.FEQ)
 If Not ErrorCode()
   Delete(Self.Controls)
 End


!-----------------------------------------------------------------------------------------------
! Register a control into the queue of controls to be checked
!-----------------------------------------------------------------------------------------------
ControlMonitor.RegisterControl  Procedure(Long pFEQ)

 Code
 Self.Controls.FEQ      = pFEQ                            !Set the FEQ
 Self.Controls.ContCRC = Self.GetCRC( pFEQ )              !Get the CRC
 Self.Controls.FEQLabel = ControlMonitorFieldName( pFEQ ) !Get the field label for the debugger
 Add(Self.Controls,Self.Controls.FEQ)





